class UnknownChord(Exception):
    def __init__(self,root_note: str,chord:str):
        self.message = f"""
        {root_note}{chord} is not a valid chord for now
        """
    
    def __str__(self):
        return self.message


import pytest
from pysongtool import PySongTool
from pysongtool.exceptions.UnknownChord import UnknownChord
from pysongtool.exceptions.WrongNote import WrongNote
from pysongtool.exceptions.UnknownScale import WrongScale

@pytest.fixture
def pysongtool_instance():
    return PySongTool()

def test_chord_valid(pysongtool_instance):
    result = pysongtool_instance.chord('C', 'maj')
    expected_notes = ['C', 'E', 'G']
    assert result['notes'] == expected_notes
    assert result['chord'] == 'Cmaj'

def test_chord_invalid_chord(pysongtool_instance):
    with pytest.raises(UnknownChord):
        pysongtool_instance.chord('C', 'unknown')

def test_chord_invalid_note(pysongtool_instance):
    with pytest.raises(WrongNote):
        pysongtool_instance.chord('H', 'maj')

def test_all_chords(pysongtool_instance):
    result = pysongtool_instance.all_chords('C')
    assert len(result) > 0
    assert 'maj' in result[0]

def test_scale_valid(pysongtool_instance):
    result = pysongtool_instance.scale('C', 'major')
    expected_notes = ['C', 'D', 'E', 'F', 'G', 'A', 'B']
    assert result['notes'] == expected_notes

def test_scale_invalid_scale(pysongtool_instance):
    with pytest.raises(WrongScale):
        pysongtool_instance.scale('C', 'unknown')

def test_scale_invalid_note(pysongtool_instance):
    with pytest.raises(WrongNote):
        pysongtool_instance.scale('H', 'major')

def test_intervals(pysongtool_instance):
    result = pysongtool_instance.intervals('C')
    assert result[0]['note'] == 'C'
    assert result[0]['semitones'] == 0

def test_intervals_invalid_note(pysongtool_instance):
    with pytest.raises(WrongNote):
        pysongtool_instance.intervals('H')

def test_get_interval(pysongtool_instance):
    result = pysongtool_instance.get_intervals('C', 'E', 'G')
    assert result[0]['note'] == 'E'
    assert result[0]['interval']['semitones'] == 4

def test_get_interval_invalid_note(pysongtool_instance):
    with pytest.raises(WrongNote):
        pysongtool_instance.get_intervals('C', 'H')



class WrongScale(Exception):
    def __init__(self,root_note: str, scale_name: str):
        self.message = f"""
        {root_note}{scale_name} is not a valid scale\n
        use only standard and sharp notes (e.g. C,C#)
        """
    
    def __str__(self):
        return self.message


from pysongtool.objects.CircularLinkedList import CircularLinkedList

from pysongtool.data.scales import scales_list
from pysongtool.data.chords import chord_list
from pysongtool.data.intervals import intervals_list
from pysongtool.data.fifths import fifths
from pysongtool.data.progressions import progression_list

from pysongtool.exceptions.UnknownChord import UnknownChord
from pysongtool.exceptions.WrongNote import WrongNote
from pysongtool.exceptions.UnknownScale import WrongScale

class PySongTool:
    """ This is the facade class who give easy acess for all pysongtool features."""

    def __init__(self, circular_linked_list_object=CircularLinkedList):
        
        self.list = circular_linked_list_object()
        self.chords = chord_list

        self.list.append('C')
        self.list.append('C#')
        self.list.append('D')
        self.list.append('D#')
        self.list.append('E')
        self.list.append('F')
        self.list.append('F#')
        self.list.append('G')
        self.list.append('G#')
        self.list.append('A')
        self.list.append('A#')
        self.list.append('B')

        self.scales_list = scales_list
        self.chords_list = chord_list
        self.intervals_list = intervals_list
        self.fifths_list = fifths
    
    def chord(self, root_note: str,chord_name: str):
        """ Use it if you want to get a chord using root_note as key"""

        root_note = root_note.upper()

        try:
            intervals = chord_list[chord_name]
            
        except:
            raise UnknownChord(root_note, chord_name)

        notes = []

        #get info about the root note (symbol,index)
        try:
            root_info: dict = self.list.find_one(root_note)
        except:
            raise WrongNote(root_note)

        #get the symbol
        notes.append(root_info[0].data)

        for i in intervals:

            #find a note using: root index + chord note interval (in semitones)
            notes.append(self.list[root_info[1] + i].data)

        return {
            'chord': f'{root_note}{chord_name}',
            'notes': notes
        }
    
    def all_chords(self, root_note: str):
        try:
            root_note = root_note.upper()
            root_info: dict = self.list.find_one(root_note)
        except:
            raise WrongNote(root_note)

        all_chords = []

        i = 1
        for c in chord_list:
            _current_chord = self.chord(root_note, c)

            all_chords.append(
                {
                    f'{c}': _current_chord
                }
            )

        return all_chords

    def scale(self, root_note: str, scale_name: str):
        """ Use it if you want to get an musical scale with root_note as a key"""

        root_note = root_note.upper()
        notes = []
        chords = []

        try:
            scale_info = scales_list[scale_name]

            intervals = scale_info['notes']
            _chords = scale_info['chords']
            
        except:
            raise WrongScale(root_note, scale_name)
        
        try:
            root_info: dict = self.list.find_one(root_note)
        except:
            raise WrongNote(root_note)

        #get the symbol
        notes.append(root_info[0].data)

        for i in intervals:

            #find a note using: root index + chord note interval (in semitones)
            notes.append(self.list[root_info[1] + i].data)

        for i in range(len(notes)):
            _current_note = notes[i]
            _current_chord = _chords[i]

            chords.append(f'{_current_note}{_current_chord}')

        return {
            'notes': notes,
            'chords': chords
        }
    
    def intervals(self, root_note: str):
        """ Use it if you want to get a dict containing all intervals with root_note as reference"""

        root_note = root_note.upper()
        notes = []

        try:
            root_info: dict = self.list.find_one(root_note)
        except:
            raise WrongNote(root_note)
        
        root = root_info[1]

        n = 0
        for i in intervals_list:
            _current_interval = i
            _semitones = _current_interval['semitones']


            _current_interval['note'] = self.list[root + _semitones].data

            notes.append(_current_interval)
            n += 1

        return notes
    
    def get_intervals(self, root_note, *args):
        root_note = root_note.upper()

        try:
            root_info: dict = self.list.find_one(root_note)
        except:
            raise WrongNote(root_note)
        
        note_and_intervals = []

        for i in range(len(args)):
            try:
                _current_note = args[i].upper()
                note: dict = self.list.find_one(_current_note)
            except:
                raise WrongNote(args[i])

            d = abs(0 - note[1])

            note_and_intervals.append(
                {
                    'note': _current_note,
                    'interval': intervals_list[d]
                }
            )

        return note_and_intervals

    def get_fifths(self):
        return self.fifths_list
    
    def all_progressions(self):
        return progression_list


progression_list = [
    {
        'degrees_repr': ['I', 'V', 'vi', 'IV'],
        'degress_halftones': [],
        'scale': 'major',
        'examples': {
            'The Beatles': [
                'Let it Be'
            ]
        }
    },

    {
        'degrees_repr': ['i', 'VII', 'VI'],
        'degress_halftones': [],
        'scale': 'minor',
        'examples': {
            'Led Zeppelin': [
                'Stairway to Heaven'
            ]
        }
    }
]



#Used for the list of notes

class Node:
    def __init__(self, data, bits=16):
        self.data = data
        self.next = None

    def __str__(self):
        return str(self.data)

""" This list only use things I need"""

class CircularLinkedList:
    def __init__(self, node_class=Node):
        self.root = None

    def append(self, data):
        new_node = Node(data)

        if not self.root:
            self.root = new_node
            self.root.next = self.root

        else:
            current_node = self.root

            while current_node.next != self.root:
                current_node = current_node.next
            
            current_node.next = new_node
            new_node.next = self.root
    
    def find_one(self, data):
        current_node = self.root
        index = 0

        while current_node.data != data:
            current_node = current_node.next
            index += 1

            if current_node == self.root:
                raise StopIteration

        return [current_node, index]
    
    def __getitem__(self, steps):
        current_node = self.root
        i = 0

        while i <= steps:
            node = current_node
            current_node = current_node.next

            i += 1

        return node
    
    def __iter__(self):
        self._iter_current_node = self.root
        return self

    def __next__(self):
        if self._iter_current_node.next != self.root:
            node = self._iter_current_node
            self._iter_current_node = self._iter_current_node.next

            return node
        
        raise StopIteration

    def __repr__(self):
        object_list = []

        current_node = self.root
        while True:
            object_list.append(str(current_node))
            print(current_node)

            current_node = current_node.next
            if current_node == self.root:
                break

        return f'Circular{str(object_list)}'

